@@ -97,6 +97,50 @@ pad_button_set_down(struct pad_dispatch *pad,
 	}
 }
 
+static void
+pad_process_relative(struct pad_dispatch *pad,
+		     struct evdev_device *device,
+		     struct input_event *e,
+		     uint64_t time)
+{
+	switch (e->code) {
+	case REL_DIAL:
+		pad->dials.dial1 = e->value * 120;
+		pad->changed_axes |= PAD_AXIS_DIAL1;
+		pad_set_status(pad, PAD_AXES_UPDATED);
+		break;
+	case REL_WHEEL:
+		if (!pad->dials.has_hires_dial) {
+			pad->dials.dial1 = e->value * 120;
+			pad->changed_axes |= PAD_AXIS_DIAL1;
+			pad_set_status(pad, PAD_AXES_UPDATED);
+		}
+		break;
+	case REL_HWHEEL:
+		if (!pad->dials.has_hires_dial) {
+			pad->dials.dial2 = e->value * 120;
+			pad->changed_axes |= PAD_AXIS_DIAL2;
+			pad_set_status(pad, PAD_AXES_UPDATED);
+		}
+		break;
+	case REL_WHEEL_HI_RES:
+		pad->dials.dial1 = e->value;
+		pad->changed_axes |= PAD_AXIS_DIAL1;
+		pad_set_status(pad, PAD_AXES_UPDATED);
+		break;
+	case REL_HWHEEL_HI_RES:
+		pad->dials.dial2 = e->value * 120;
+		pad->changed_axes |= PAD_AXIS_DIAL2;
+		pad_set_status(pad, PAD_AXES_UPDATED);
+		break;
+	default:
+		evdev_log_info(device,
+			       "Unhandled EV_REL event code %#x\n",
+			       e->code);
+		break;
+	}
+}
+
 static void
 pad_process_absolute(struct pad_dispatch *pad,
 		     struct evdev_device *device,
@@ -146,7 +190,7 @@ pad_process_absolute(struct pad_dispatch *pad,
 }
 
 static inline double
-normalize_ring(const struct input_absinfo *absinfo)
+normalize_wacom_ring(const struct input_absinfo *absinfo)
 {
 	/* libinput has 0 as the ring's northernmost point in the device's
 	   current logical rotation, increasing clockwise to 1. Wacom has
@@ -162,7 +206,7 @@ normalize_ring(const struct input_absinfo *absinfo)
 }
 
 static inline double
-normalize_strip(const struct input_absinfo *absinfo)
+normalize_wacom_strip(const struct input_absinfo *absinfo)
 {
 	/* strip axes don't use a proper value, they just shift the bit left
 	 * for each position. 0 isn't a real value either, it's only sent on
@@ -175,6 +219,12 @@ normalize_strip(const struct input_absinfo *absinfo)
 	return value;
 }
 
+static inline double
+normalize_strip(const struct input_absinfo *absinfo)
+{
+	return absinfo_normalize_value(absinfo, absinfo->value);
+}
+
 static inline double
 pad_handle_ring(struct pad_dispatch *pad,
 		struct evdev_device *device,
@@ -186,7 +236,7 @@ pad_handle_ring(struct pad_dispatch *pad,
 	absinfo = libevdev_get_abs_info(device->evdev, code);
 	assert(absinfo);
 
-	degrees = normalize_ring(absinfo) * 360;
+	degrees = normalize_wacom_ring(absinfo) * 360;
 
 	if (device->left_handed.enabled)
 		degrees = fmod(degrees + 180, 360);
@@ -208,7 +258,10 @@ pad_handle_strip(struct pad_dispatch *pad,
 	if (absinfo->value == 0)
 		return 0.0;
 
-	pos = normalize_strip(absinfo);
+	if (evdev_device_get_id_vendor(device) == VENDOR_ID_WACOM)
+		pos = normalize_wacom_strip(absinfo);
+	else
+		pos = normalize_strip(absinfo);
 
 	if (device->left_handed.enabled)
 		pos = 1.0 - pos;
@@ -216,6 +269,22 @@ pad_handle_strip(struct pad_dispatch *pad,
 	return pos;
 }
 
+static inline struct libinput_tablet_pad_mode_group *
+pad_dial_get_mode_group(struct pad_dispatch *pad,
+			unsigned int dial)
+{
+	struct libinput_tablet_pad_mode_group *group;
+
+	list_for_each(group, &pad->modes.mode_group_list, link) {
+		if (libinput_tablet_pad_mode_group_has_dial(group, dial))
+			return group;
+	}
+
+	assert(!"Unable to find dial mode group");
+
+	return NULL;
+}
+
 static inline struct libinput_tablet_pad_mode_group *
 pad_ring_get_mode_group(struct pad_dispatch *pad,
 			unsigned int ring)
@@ -264,6 +333,26 @@ pad_check_notify_axes(struct pad_dispatch *pad,
 	    libevdev_get_event_value(device->evdev, EV_ABS, ABS_MISC) == 0)
 		send_finger_up = true;
 
+	/* Unlike the ring axis we don't get an event when we release
+	 * so we can't set a source */
+	if (pad->changed_axes & PAD_AXIS_DIAL1) {
+		group = pad_dial_get_mode_group(pad, 0);
+		tablet_pad_notify_dial(base,
+				       time,
+				       0,
+				       pad->dials.dial1,
+				       group);
+	}
+
+	if (pad->changed_axes & PAD_AXIS_DIAL2) {
+		group = pad_dial_get_mode_group(pad, 1);
+		tablet_pad_notify_dial(base,
+				       time,
+				       1,
+				       pad->dials.dial2,
+				       group);
+	}
+
 	if (pad->changed_axes & PAD_AXIS_RING1) {
 		value = pad_handle_ring(pad, device, ABS_WHEEL);
 		if (send_finger_up)
@@ -473,6 +562,8 @@ pad_flush(struct pad_dispatch *pad,
 	memcpy(&pad->prev_button_state,
 	       &pad->button_state,
 	       sizeof(pad->button_state));
+	pad->dials.dial1 = 0;
+	pad->dials.dial2 = 0;
 }
 
 static void
@@ -484,6 +575,9 @@ pad_process(struct evdev_dispatch *dispatch,
 	struct pad_dispatch *pad = pad_dispatch(dispatch);
 
 	switch (e->type) {
+	case EV_REL:
+		pad_process_relative(pad, device, e, time);
+		break;
 	case EV_ABS:
 		pad_process_absolute(pad, device, e, time);
 		break;
@@ -686,6 +780,16 @@ pad_init(struct pad_dispatch *pad, struct evdev_device *device)
 	pad->status = PAD_NONE;
 	pad->changed_axes = PAD_AXIS_NONE;
 
+        /* We expect the kernel to either give us both axes as hires or neither.
+	 * Getting one is a kernel bug we don't need to care about */
+        pad->dials.has_hires_dial = libevdev_has_event_code(device->evdev, EV_REL, REL_WHEEL_HI_RES) ||
+				    libevdev_has_event_code(device->evdev, EV_REL, REL_HWHEEL_HI_RES);
+
+	if (libevdev_has_event_code(device->evdev, EV_REL, REL_WHEEL) &&
+	    libevdev_has_event_code(device->evdev, EV_REL, REL_DIAL)) {
+		log_bug_libinput(pad_libinput_context(pad), "Unsupported combination REL_DIAL and REL_WHEEL\n");
+	}
+
 	pad_init_buttons(pad, device);
 	pad_init_left_handed(device);
 	if (pad_init_leds(pad, device) != 0)
@@ -782,6 +886,26 @@ evdev_device_tablet_pad_get_num_buttons(struct evdev_device *device)
 	return pad->nbuttons;
 }
 
+int
+evdev_device_tablet_pad_get_num_dials(struct evdev_device *device)
+{
+	int ndials = 0;
+
+	if (!(device->seat_caps & EVDEV_DEVICE_TABLET_PAD))
+		return -1;
+
+	if (libevdev_has_event_code(device->evdev, EV_REL, REL_WHEEL) ||
+	    libevdev_has_event_code(device->evdev, EV_REL, REL_DIAL)) {
+		ndials++;
+		if (libevdev_has_event_code(device->evdev,
+					    EV_REL,
+					    REL_HWHEEL))
+			ndials++;
+	}
+
+	return ndials;
+}
+
 int
 evdev_device_tablet_pad_get_num_rings(struct evdev_device *device)
 {
