@@ -129,14 +129,20 @@ evdev_update_key_down_count(struct evdev_device *device,
 			    int code,
 			    int pressed)
 {
-	int key_count;
+	int key_count = 0;
 	assert(code >= 0 && code < KEY_CNT);
 
 	if (pressed) {
 		key_count = ++device->key_count[code];
 	} else {
-		assert(device->key_count[code] > 0);
-		key_count = --device->key_count[code];
+		if (device->key_count[code] > 0) {
+			key_count = --device->key_count[code];
+		} else {
+			evdev_log_bug_libinput(device,
+					       "releasing key %s with count %d\n",
+					       libevdev_event_code_get_name(EV_KEY, code),
+					       device->key_count[code]);
+		}
 	}
 
 	if (key_count > 32) {
@@ -332,6 +338,8 @@ evdev_device_led_update(struct evdev_device *device, enum libinput_led leds)
 		{ LIBINPUT_LED_NUM_LOCK, LED_NUML },
 		{ LIBINPUT_LED_CAPS_LOCK, LED_CAPSL },
 		{ LIBINPUT_LED_SCROLL_LOCK, LED_SCROLLL },
+		{ LIBINPUT_LED_COMPOSE, LED_COMPOSE },
+		{ LIBINPUT_LED_KANA, LED_KANA },
 	};
 	struct input_event ev[ARRAY_LENGTH(map) + 1];
 	unsigned int i;
@@ -378,8 +386,7 @@ evdev_transform_relative(struct evdev_device *device,
 static inline double
 scale_axis(const struct input_absinfo *absinfo, double val, double to_range)
 {
-	return (val - absinfo->minimum) * to_range /
-		(absinfo->maximum - absinfo->minimum + 1);
+	return (val - absinfo->minimum) * to_range / absinfo_range(absinfo);
 }
 
 double
@@ -957,8 +964,7 @@ evdev_scroll_config_natural_get(struct libinput_device *device)
 static int
 evdev_scroll_config_natural_get_default(struct libinput_device *device)
 {
-	/* could enable this on Apple touchpads. could do that, could
-	 * very well do that... */
+	/* Overridden in evdev-mt-touchpad.c for Apple touchpads. */
 	return 0;
 }
 
@@ -1416,7 +1422,7 @@ evdev_read_wheel_click_props(struct evdev_device *device)
 	if (evdev_read_wheel_click_count_prop(device, hwheel_count, &angles.x) ||
 	    evdev_read_wheel_click_prop(device, hwheel_angle, &angles.x)) {
 		evdev_log_debug(device,
-				"wheel: horizontal click angle: %.2f\n", angles.y);
+				"wheel: horizontal click angle: %.2f\n", angles.x);
 	} else {
 		angles.x = angles.y;
 	}
@@ -1674,8 +1680,8 @@ evdev_fix_abs_resolution(struct evdev_device *device,
 	 */
 	if (!evdev_read_attr_res_prop(device, &xres, &yres) &&
 	    evdev_read_attr_size_prop(device, &widthmm, &heightmm)) {
-		xres = (absx->maximum - absx->minimum)/widthmm;
-		yres = (absy->maximum - absy->minimum)/heightmm;
+		xres = absinfo_range(absx)/widthmm;
+		yres = absinfo_range(absy)/heightmm;
 	}
 
 	/* libevdev_set_abs_resolution() changes the absinfo we already
@@ -1844,10 +1850,8 @@ evdev_extract_abs_axes(struct evdev_device *device,
 
 	device->abs.absinfo_x = libevdev_get_abs_info(evdev, ABS_X);
 	device->abs.absinfo_y = libevdev_get_abs_info(evdev, ABS_Y);
-	device->abs.dimensions.x = abs(device->abs.absinfo_x->maximum -
-				       device->abs.absinfo_x->minimum);
-	device->abs.dimensions.y = abs(device->abs.absinfo_y->maximum -
-				       device->abs.absinfo_y->minimum);
+	device->abs.dimensions.x = abs((int)absinfo_range(device->abs.absinfo_x));
+	device->abs.dimensions.y = abs((int)absinfo_range(device->abs.absinfo_y));
 
 	if (evdev_is_fake_mt_device(device) ||
 	    !libevdev_has_event_code(evdev, EV_ABS, ABS_MT_POSITION_X) ||
@@ -1866,10 +1870,8 @@ evdev_extract_abs_axes(struct evdev_device *device,
 
 	device->abs.absinfo_x = libevdev_get_abs_info(evdev, ABS_MT_POSITION_X);
 	device->abs.absinfo_y = libevdev_get_abs_info(evdev, ABS_MT_POSITION_Y);
-	device->abs.dimensions.x = abs(device->abs.absinfo_x->maximum -
-				       device->abs.absinfo_x->minimum);
-	device->abs.dimensions.y = abs(device->abs.absinfo_y->maximum -
-				       device->abs.absinfo_y->minimum);
+	device->abs.dimensions.x = abs((int)absinfo_range(device->abs.absinfo_x));
+	device->abs.dimensions.y = abs((int)absinfo_range(device->abs.absinfo_y));
 	device->is_mt = 1;
 }
 
@@ -2517,6 +2519,12 @@ evdev_device_get_name(struct evdev_device *device)
 	return device->devname;
 }
 
+unsigned int
+evdev_device_get_id_bustype(struct evdev_device *device)
+{
+	return libevdev_get_id_bustype(device->evdev);
+}
+
 unsigned int
 evdev_device_get_id_product(struct evdev_device *device)
 {
@@ -2563,8 +2571,8 @@ evdev_device_calibrate(struct evdev_device *device,
 		return;
 	}
 
-	sx = device->abs.absinfo_x->maximum - device->abs.absinfo_x->minimum + 1;
-	sy = device->abs.absinfo_y->maximum - device->abs.absinfo_y->minimum + 1;
+	sx = absinfo_range(device->abs.absinfo_x);
+	sy = absinfo_range(device->abs.absinfo_y);
 
 	/* The transformation matrix is in the form:
 	 *  [ a b c ]
@@ -2717,6 +2725,10 @@ evdev_device_get_size(const struct evdev_device *device,
 	x = libevdev_get_abs_info(device->evdev, ABS_X);
 	y = libevdev_get_abs_info(device->evdev, ABS_Y);
 
+	if ((x && x->minimum == 0 && x->maximum == 1) ||
+	    (y && y->minimum == 0 && y->maximum == 1))
+		return -1;
+
 	if (!x || !y || device->abs.is_fake_resolution ||
 	    !x->resolution || !y->resolution)
 		return -1;
@@ -3120,7 +3132,7 @@ evdev_device_destroy(struct evdev_device *device)
 bool
 evdev_tablet_has_left_handed(struct evdev_device *device)
 {
-	bool has_left_handed = false;
+	bool has_left_handed = true;
 #if HAVE_LIBWACOM
 	struct libinput *li = evdev_libinput_context(device);
 	WacomDeviceDatabase *db = NULL;
@@ -3141,8 +3153,7 @@ evdev_tablet_has_left_handed(struct evdev_device *device)
 				   error);
 
 	if (d) {
-		if (libwacom_is_reversible(d))
-			has_left_handed = true;
+		has_left_handed = !!libwacom_is_reversible(d);
 	} else if (libwacom_error_get_code(error) == WERROR_UNKNOWN_MODEL) {
 		evdev_log_info(device,
 			       "tablet '%s' unknown to libwacom\n",
