@@ -473,6 +473,9 @@ evdev_device_get_sysname(struct evdev_device *device);
 const char *
 evdev_device_get_name(struct evdev_device *device);
 
+unsigned int
+evdev_device_get_id_bustype(struct evdev_device *device);
+
 unsigned int
 evdev_device_get_id_product(struct evdev_device *device);
 
@@ -518,6 +521,9 @@ evdev_device_tablet_pad_has_key(struct evdev_device *device,
 int
 evdev_device_tablet_pad_get_num_buttons(struct evdev_device *device);
 
+int
+evdev_device_tablet_pad_get_num_dials(struct evdev_device *device);
+
 int
 evdev_device_tablet_pad_get_num_rings(struct evdev_device *device);
 
