@@ -158,6 +158,9 @@ print_event_header(struct libinput_event *ev)
 	case LIBINPUT_EVENT_TABLET_PAD_KEY:
 		type = "TABLET_PAD_KEY";
 		break;
+	case LIBINPUT_EVENT_TABLET_PAD_DIAL:
+		type = "TABLET_PAD_DIAL";
+		break;
 	case LIBINPUT_EVENT_SWITCH_TOGGLE:
 		type = "SWITCH_TOGGLE";
 		break;
@@ -494,7 +497,7 @@ print_pointer_axis_event(struct libinput_event *ev)
 	double v = 0, h = 0, v120 = 0, h120 = 0;
 	const char *have_vert = "",
 		   *have_horiz = "";
-	const char *source = "invalid";
+	const char *source = NULL;
 	enum libinput_pointer_axis axis;
 	enum libinput_event_type type;
 
@@ -746,7 +749,7 @@ static void
 print_tablet_pad_ring_event(struct libinput_event *ev)
 {
 	struct libinput_event_tablet_pad *p = libinput_event_get_tablet_pad_event(ev);
-	const char *source = "<invalid>";
+	const char *source = NULL;
 	unsigned int mode;
 
 	print_event_time(libinput_event_tablet_pad_get_time(p));
@@ -772,7 +775,7 @@ static void
 print_tablet_pad_strip_event(struct libinput_event *ev)
 {
 	struct libinput_event_tablet_pad *p = libinput_event_get_tablet_pad_event(ev);
-	const char *source = "<invalid>";
+	const char *source = NULL;
 	unsigned int mode;
 
 	print_event_time(libinput_event_tablet_pad_get_time(p));
@@ -819,6 +822,21 @@ print_tablet_pad_key_event(struct libinput_event *ev)
 	       state == LIBINPUT_KEY_STATE_PRESSED ? "pressed" : "released");
 }
 
+static void
+print_tablet_pad_dial_event(struct libinput_event *ev)
+{
+	struct libinput_event_tablet_pad *p = libinput_event_get_tablet_pad_event(ev);
+	unsigned int mode;
+
+	print_event_time(libinput_event_tablet_pad_get_time(p));
+
+	mode = libinput_event_tablet_pad_get_mode(p);
+	printq("dial %d delta %.2f (mode %d)\n",
+	       libinput_event_tablet_pad_get_dial_number(p),
+	       libinput_event_tablet_pad_get_dial_delta_v120(p),
+	       mode);
+}
+
 static void
 print_switch_event(struct libinput_event *ev)
 {
@@ -943,6 +961,9 @@ handle_and_print_events(struct libinput *li)
 		case LIBINPUT_EVENT_TABLET_PAD_KEY:
 			print_tablet_pad_key_event(ev);
 			break;
+		case LIBINPUT_EVENT_TABLET_PAD_DIAL:
+			print_tablet_pad_dial_event(ev);
+			break;
 		case LIBINPUT_EVENT_SWITCH_TOGGLE:
 			print_switch_event(ev);
 			break;
