@@ -190,6 +190,10 @@ struct window {
 			double position;
 			int number;
 		} strip;
+		struct {
+			double position;
+			int number;
+		} dial;
 	} pad;
 
 	struct {
@@ -234,7 +238,7 @@ wayland_registry_global_remove(void *data,
 
 }
 
-struct wl_registry_listener registry_listener = {
+static struct wl_registry_listener registry_listener = {
 	wayland_registry_global,
 	wayland_registry_global_remove
 };
@@ -448,7 +452,6 @@ draw_evdev_abs(struct window *w, cairo_t *cr)
 	int x, y;
 
 	cairo_save(cr);
-	cairo_set_source_rgb(cr, .2, .2, .8);
 
 	center_x = w->width/2 + 400;
 	center_y = w->height/2;
@@ -495,6 +498,7 @@ draw_evdev_abs(struct window *w, cairo_t *cr)
 		if (!w->evdev.slots[i].active)
 			continue;
 
+		cairo_set_source_rgb(cr, .2, .2, .8);
 		x = w->evdev.slots[i].x;
 		y = w->evdev.slots[i].y;
 		x = 1.0 * (x - ax->minimum)/width * outline_width;
@@ -503,10 +507,21 @@ draw_evdev_abs(struct window *w, cairo_t *cr)
 		y += center_y - outline_height/2;
 		cairo_arc(cr, x, y, 10, 0, 2 * M_PI);
 		cairo_fill(cr);
+
+		char finger_text[3];
+		cairo_text_extents_t finger_text_extents;
+		snprintf(finger_text, 3, "%zu", i);
+		cairo_set_source_rgb(cr, 1.f, 1.f, 1.f);
+		cairo_set_font_size(cr, 12.0);
+		cairo_text_extents(cr, finger_text, &finger_text_extents);
+		cairo_move_to(cr, x - finger_text_extents.width/2,
+				  y + finger_text_extents.height/2);
+		cairo_show_text(cr, finger_text);
 	}
 
 draw_outline:
 	/* The touchpad outline */
+	cairo_set_source_rgb(cr, .2, .2, .8);
 	cairo_rectangle(cr,
 			center_x - outline_width/2,
 			center_y - outline_height/2,
@@ -709,12 +724,12 @@ draw_pad(struct window *w, cairo_t *cr)
 	ry = w->height/2 + 100;
 
 	cairo_save(cr);
-	/* outer ring */
+	/* outer ring (for ring) */
 	cairo_set_source_rgb(cr, .7, .7, .0);
 	cairo_arc(cr, rx, ry, 50, 0, 2 * M_PI);
 	cairo_fill(cr);
 
-	/* inner ring */
+	/* inner ring (for dial) */
 	cairo_set_source_rgb(cr, 1., 1., 1.);
 	cairo_arc(cr, rx, ry, 30, 0, 2 * M_PI);
 	cairo_fill(cr);
@@ -732,7 +747,20 @@ draw_pad(struct window *w, cairo_t *cr)
 		snprintf(number, sizeof(number), "%d", w->pad.ring.number);
 		cairo_set_source_rgb(cr, .0, .0, .0);
 		draw_text(cr, number, rx, ry);
+	}
 
+	if (w->pad.dial.position != -1) {
+		const int degrees_per_click = 15.0;
+		double degrees = fmod(w->pad.dial.position/120 * degrees_per_click, 360);
+		pos = (degrees + 270) * M_PI/180.0;
+		cairo_set_source_rgb(cr, .0, .0, .0);
+		cairo_set_line_width(cr, 20);
+		cairo_arc(cr, rx, ry, 20, pos - M_PI/12 , pos + M_PI/12);
+		cairo_stroke(cr);
+
+		snprintf(number, sizeof(number), "%d", w->pad.dial.number);
+		cairo_set_source_rgb(cr, .0, .0, .0);
+		draw_text(cr, number, rx, ry);
 	}
 
 	cairo_restore(cr);
@@ -970,6 +998,7 @@ draw(GtkWidget *widget, cairo_t *cr, gpointer data)
 {
 	struct window *w = data;
 
+	cairo_set_font_size(cr, 12.0);
 	cairo_set_source_rgb(cr, 1, 1, 1);
 	cairo_rectangle(cr, 0, 0, w->width, w->height);
 	cairo_fill(cr);
@@ -1134,7 +1163,7 @@ window_init(struct window *w)
 	g_signal_connect(G_OBJECT(w->win), "close-request", G_CALLBACK(window_delete_event_cb), w);
 
 	gtk_window_set_child(GTK_WINDOW(w->win), w->area);
-	gtk_widget_show(w->win);
+	gtk_widget_set_visible(w->win, TRUE);
 #else
 	g_signal_connect(G_OBJECT(w->win), "map-event", G_CALLBACK(map_event_cb), w);
 	g_signal_connect(G_OBJECT(w->win), "delete-event", G_CALLBACK(window_delete_event_cb), w);
@@ -1148,6 +1177,7 @@ window_init(struct window *w)
 
 	w->pad.ring.position = -1;
 	w->pad.strip.position = -1;
+	w->pad.dial.position = -1;
 }
 
 static void
@@ -1714,7 +1744,7 @@ handle_event_tablet_pad(struct libinput_event *ev, struct window *w)
 		"Pad 0", "Pad 1", "Pad 2", "Pad 3", "Pad 4", "Pad 5",
 		"Pad 6", "Pad 7", "Pad 8", "Pad 9", "Pad >= 10"
 	};
-	double position;
+	double position, delta;
 	double number;
 
 	switch (libinput_event_get_type(ev)) {
@@ -1736,6 +1766,14 @@ handle_event_tablet_pad(struct libinput_event *ev, struct window *w)
 		w->pad.strip.number = number;
 		w->pad.strip.position = position;
 		break;
+	case LIBINPUT_EVENT_TABLET_PAD_DIAL:
+		delta = libinput_event_tablet_pad_get_dial_delta_v120(p);
+		number = libinput_event_tablet_pad_get_dial_number(p);
+		if (w->pad.dial.number != number)
+			w->pad.dial.position = -delta;
+		w->pad.dial.number = number;
+		w->pad.dial.position += delta;
+		break;
 	default:
 		abort();
 	}
@@ -1813,6 +1851,7 @@ handle_event_libinput(GIOChannel *source, GIOCondition condition, gpointer data)
 		case LIBINPUT_EVENT_TABLET_PAD_BUTTON:
 		case LIBINPUT_EVENT_TABLET_PAD_RING:
 		case LIBINPUT_EVENT_TABLET_PAD_STRIP:
+		case LIBINPUT_EVENT_TABLET_PAD_DIAL:
 			handle_event_tablet_pad(ev, w);
 			break;
 		case LIBINPUT_EVENT_TABLET_PAD_KEY:
